#include <autoconf.h>
#include <hermit.h>
#include <herrno.h>
#include <init.h>
#include <platform.h>
#include <flash.h>
#include <rcm2_gpio.h>
#include <rcm2_sdhc.h>
#include <m24c01.h>

static void adjust_recovery_boot_args(struct platform_info *pinfo,
				      int *argc, char *argv[]);
static void adjust_recovery_boot_device(struct platform_info *pinfo,
					char *device);
static void adjust_recovery_fdt_device(struct platform_info *pinfo,
					char *device);

static struct page_table pt_list[] = {
	/* Physical,  Virtual,    Size,       Option */
	/*-------------------------------------------*/
	/* SDRAM */
	{0x40000000, 0x40000000, 0x00b00000, 0xc0e},
	{0x40b00000, 0x40b00000, 0x00100000, 0xc02},
	{0x40c00000, 0x40c00000, 0x7f400000, 0xc0e},

	/* System-Domain Peripheral bus */
	{0xe6000000, 0xe6000000, 0x02000000, 0xc02},
	/* GIC */
	{0xf1000000, 0xf1000000, 0x00100000, 0xc02},
	/* SDHI */
	{0xee100000, 0xee100000, 0x00100000, 0xc02},
	/* eMMC */
	{0xee200000, 0xee200000, 0x00100000, 0xc02},
	/* EtherMAC */
	{0xee700000, 0xee700000, 0x00200000, 0xc02},
};

#include "board.c"

int rcm2_uart_request_port(int port)
{
	switch (port) {
	case 2:
		modsel(MOD_SCIFB, MOD_FUNC1);
		pinmux(xSCIFB0_RXD, GP_FUNC, IP_FUNC2, PU_EN);
		pinmux(xSCIFB0_TXD, GP_FUNC, IP_FUNC2, PU_DIS);
		break;
	case 3:
		modsel(MOD_SCIFB1, MOD_FUNC1);
		pinmux(xSCIFB1_RXD, GP_FUNC, IP_FUNC2, PU_EN);
		pinmux(xSCIFB1_TXD, GP_FUNC, IP_FUNC2, PU_DIS);
		break;
	default:
		return -H_EINVAL;
	}

	return 0;
}

int rcm2_qspi_request_port(void)
{
	modsel(MOD_QSP, MOD_FUNC1);

	pinmux(xSSL, GP_FUNC, IP_FUNC3, PU_DIS);
	pinmux(xSPCLK, GP_FUNC, IP_FUNC2, PU_EN);
	pinmux(xMOSI_IO0, GP_FUNC, IP_FUNC3, PU_EN);
	pinmux(xMISO_IO1, GP_FUNC, IP_FUNC2, PU_EN);
	pinmux(xIO2, GP_FUNC, IP_FUNC2, PU_EN);
	pinmux(xIO3, GP_FUNC, IP_FUNC3, PU_EN);

	return 0;
}

int rcm2_mmc_request_port(void)
{
	modsel(MOD_MMC, MOD_FUNC1);

	pinmux(xMMC_CLK, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_CMD, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_D0, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_D1, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_D2, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_D3, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_D4, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_D5, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_D6, GP_FUNC, IP_FUNC2, PU_DIS);
	pinmux(xMMC_D7, GP_FUNC, IP_FUNC2, PU_DIS);

	return 0;
}

int rcm2_sdhc_request_port(int port)
{
	switch (port) {
	case 0:
		pinmux(xSD0_CLK, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD0_CMD, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD0_DATA0, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD0_DATA1, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD0_DATA2, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD0_DATA3, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD0_CD, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD0_WP, GP_FUNC, IP_FUNC1, PU_DIS);

		pinmux(xSD0_PWR_EN, GP_GPIO, IP_FUNC1, PU_EN);
		break;
	case 1:
		pinmux(xSD1_CLK, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD1_CMD, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD1_DATA0, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD1_DATA1, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD1_DATA2, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD1_DATA3, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD1_CD, GP_FUNC, IP_FUNC1, PU_DIS);
		pinmux(xSD1_WP, GP_FUNC, IP_FUNC1, PU_DIS);

		pinmux(xSD1_PWR_EN, GP_GPIO, IP_FUNC1, PU_EN);
		break;
	default:
		return -H_EINVAL;
	}

	return 0;
}

int rcm2_eth_request_port(void)
{
	pinmux(xETH_REFCLK, GP_FUNC, IP_FUNC1, PU_EN);
	pinmux(xETH_TXD0, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_TXD1, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_TX_EN, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_CRS_DV, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_RX_ER, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_RXD0, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_RXD1, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_MDC, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_MDIO, GP_FUNC, IP_FUNC1, PU_DIS);
	pinmux(xETH_LINK, GP_FUNC, IP_FUNC1, PU_DIS);

	pinmux(xETH_RESET, GP_GPIO, IP_FUNC1, PU_DIS);

	return 0;
}

static void armadillo100eva_setup_flash(struct platform_info *pinfo)
{
	int val;

	flash_initialize(FLASH_TYPE_SPI, 0);

	val = flash_get_size(0);
	pinfo->map->flash.base = 0;
	pinfo->map->flash.size = 1 << val;
	hsprintf(pinfo->default_mtdparts,
		 "physmap-flash.0:"
		 "%p(all)ro,"
		 "0x10000@0(ipl)ro,"
		 "0x80000(bootloader)ro,"
                 "0x40000(fdt),"
                 "-(firmware)ro",
		 pinfo->map->flash.size);
}

static void adjust_recovery_boot_args(struct platform_info *pinfo,
				      int *argc, char *argv[])
{
	*argc = 6;
	argv[1] = "console=ttySC3,115200";
	argv[2] = "mem=1024M";
	argv[3] = "noinitrd";
	argv[4] = "rootwait";
	argv[5] = "root=/dev/mmcblk1p2";
}

static void adjust_recovery_boot_device(struct platform_info *pinfo,
					char *device)
{
	hsnprintf(device, CMDLINE_MAXLEN, "mmcblk1p2");
}

static void adjust_recovery_fdt_device(struct platform_info *pinfo,
					char *device)
{
	hsnprintf(device, CMDLINE_MAXLEN, "mmcblk1p1");
}

static void armadilloeva1500_setup_map(struct platform_info *pinfo)
{
	size_t size;

	size = DRAM_SIZE;
	pinfo->map->ram.size = size;
	hsprintf(pinfo->default_ramparts, "%p(dram-1)", size);
}

static void armadilloeva1500_display_info(struct platform_info *pinfo)
{
	struct board_private *priv = pinfo->private_data;

	hprintf("   Board Type: 0x%x\n"
		"     Revision: 0x%x\n"
		"          Lot: 0x%x\n"
		"Serial Number: %d\n"
		"    Boot Mode: %s\n",
		priv->hw_type, priv->product_rev_major,
		priv->lot_num, priv->serial_num,
		get_boot_mode_str(priv->boot_mode));

	hprintf("   ORIG MAC-1: ");
	print_mac(priv->mac1);
	if (is_valid_mac(priv->umac1)) {
		hprintf("   USER MAC-1: ");
		print_mac(priv->umac1);
	}

	hprintf("      Tact-SW: %s,%s,%s,%s\n",
		on_off_string(priv->tactsw[0]),
		on_off_string(priv->tactsw[1]),
		on_off_string(priv->tactsw[2]),
		on_off_string(priv->tactsw[3]));
}

static void armadilloeva1500_setup_private_data(struct platform_info *pinfo)
{
	struct board_private *priv = pinfo->private_data;
	u8 val;

	m24c01_read(0x00, priv->mac1, 6);
	m24c01_read(0x40, priv->umac1, 6);
	m24c01_read(0x20, (u8 *)&priv->hw_type, 4);
	m24c01_read(0x24, (u8 *)&priv->product_rev_major, 2);
	m24c01_read(0x26, (u8 *)&priv->product_rev_minor, 2);
	m24c01_read(0x2c, (u8 *)&priv->lot_num, 3);
	m24c01_read(0x2f, &val, 1);
	if (val)
		m24c01_read(0x28, (u8 *)&priv->serial_num, 4);

	pinfo->system_rev = ((u32)(priv->product_rev_major << 16) |
			     (u32)priv->lot_num);
	pinfo->serial_num = priv->serial_num;
	pinfo->machine_nr = 0;

	priv->boot_mode = (read32(RST_MODEMR)) & 0x0e;

	priv->jumper = gpio_get_value(xSW2_8) ? 0 : 1;

	priv->tactsw[0] = gpio_get_value(xSW3) ? 0 : 1;
	priv->tactsw[1] = gpio_get_value(xSW4) ? 0 : 1;
	priv->tactsw[2] = gpio_get_value(xSW5) ? 0 : 1;
	priv->tactsw[3] = gpio_get_value(xSW6) ? 0 : 1;
}

static void armadilloeva1500_init(void)
{
	struct platform_info *pinfo = &platform_info;

	armadilloeva15x0_pre_common_init();

	m24c01_request_port();
	rcm2_uart_request_port(3);
	rcm2_qspi_request_port();
	rcm2_mmc_request_port();
	rcm2_sdhc_request_port(0);
	rcm2_sdhc_request_port(1);
	rcm2_eth_request_port();

	/* LED4/5 ON, LED6/7 OFF */
	gpio_direction_output(xLED4, 1);
	gpio_direction_output(xLED5, 1);
	gpio_direction_output(xLED6, 0);
	gpio_direction_output(xLED7, 0);

	/* setting changeover switch IC2/IC3 */
	gpio_direction_output(xMD_SEL, 0);

	armadilloeva1500_setup_private_data(pinfo);
	update_target_profile(pinfo);

	memset((void *)RAM_ADDR(0x05000000), 0, 0x00020000);

	armadillo100eva_setup_flash(pinfo);
	armadilloeva1500_setup_map(pinfo);

	armadilloeva15x0_setup_ethernet(pinfo);

	memdev_add(flash, &mdev_param,
		   "hermit/param", 0x00050000, 0x00008000);

	armadilloeva15x0_post_common_init();

}
arch_initcall(armadilloeva1500_init);

static void armadilloeva1500_late_init(void)
{
	struct platform_info *pinfo = &platform_info;
	armadilloeva15x0_setup_console(pinfo);
}
latearch_initcall(armadilloeva1500_late_init);

struct platform_info platform_info = {
	.prepare_normal_boot = armadilloeva15x0_prepare_normal_boot,
	.is_autoboot = armadilloeva15x0_is_autoboot,
	.led_on = armadilloeva15x0_led_on,
	.led_off = armadilloeva15x0_led_off,
	.udelay = armadilloeva15x0_udelay,
	.get_mmu_page_tables = armadilloeva15x0_get_mmu_page_tables,
	.display_info = armadilloeva1500_display_info,
	.get_mac_address = armadilloeva15x0_get_mac_address,

	.private_data = &board_priv,
	.map = &armadilloeva15x0_memory_map,

	.net_dev = &rcm2_eth,
#if defined(CONFIG_MMCSD_RCM2_SDHC)
	.mmcsd_dev = &rcm2_mmcsd,
#endif
};
