#define CORE_NAME "rcm2_sdhc"

#include <autoconf.h>
#if defined(CONFIG_MMCSD_DEBUG)
#define TRACE_DEBUG
#define DEBUG
#endif
#include <hermit.h>
#include <herrno.h>
#include <platform.h>
#include <delay.h>
#include <arch/system.h>
#include <mmcsd.h>
#include <rcm2_gpio.h>
#include <platform/pinmux.h>
#include <da9063.h>

extern int rcm2_sdhc_request_port(int port);

static int get_sdhi_port(u32 base)
{
	int port;

	switch (base) {
	case SDHI0_BASE_ADDR:
		port = 0;
		break;
	case SDHI1_BASE_ADDR:
		port = 1;
		break;
	default:
		port = -H_EINVAL;
		break;
	}

	return port;
}

#define SDHI_READ16(base, addr) \
	read16(base + (addr << (get_sdhi_port(base) ? 0 : 1)))

#define SDHI_WRITE16(base, addr, val) \
	write16(base + (addr << (get_sdhi_port(base) ? 0 : 1)), val)

enum {
	SDHC_POWER_OFF = 0,
	SDHC_POWER_ON,
};

static int sdhc_set_pwr(int port, int state)
{
	switch (port) {
	case 0:
		gpio_direction_output(PIN_MSIOF0_SCK, state);
		break;
	case 1:
		gpio_direction_output(PIN_MSIOF0_TXD, state);
		break;
	default:
		return -H_EINVAL;
	}

	return 0;
}

enum {
	SDHC_1_8VLT = 0,
	SDHC_3_3VLT,
};

static void sdhc_clock_start(struct mmc_host *host)
{
	u16 val;
	trace();

	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));

	val = SDHI_READ16(host->base, SDHI_CLK_CTRL);
	val |= 0x0300;
	SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);

}

static void __attribute__((unused)) sdhc_clock_stop(struct mmc_host *host)
{
	u32 val;

	trace();
	/* Waiting for SD Bus busy to be cleared */
	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));
	val = SDHI_READ16(host->base, SDHI_CLK_CTRL);
	val &= ~0x0300;
	SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);
}

static int sdhc_set_clock(struct mmc_host *host, u32 clock)
{
	u32 val;
	int port;

	trace();

	/* Waiting for SD Bus busy to be cleared */
	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));

	val = SDHI_READ16(host->base, SDHI_CLK_CTRL);
	val &= ~0xff;

	port = get_sdhi_port(host->base);
	if (clock) {
		/* high speed: 24.375MHz */
		if (port == 0)
			val |= 0x01;
		else
			val |= 0x00;
	} else {
		/* low speed: 190.4kHz */
		if (port == 0)
			val |= 0x80;
		else
			val |= 0x40;
	}
	SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);

	return 0;
}

static int sdhc_pio_transfer_data(struct mmc_host *host,
				  struct mmc_command *cmd)
{
	u16 *buffer = (u16 *)cmd->data->buf;
	int i;
	u16 val;

	trace();

	while(!(SDHI_READ16(host->base, SDHI_INFO2) & 0x0100));
	if (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x0100))
		dev_err("DATA not Ready\n");

	for (i=0; i<cmd->data->blocks; i++) {
		int blksz = (int)cmd->data->blksz;
		u16 stat;

		do {
			stat = SDHI_READ16(host->base, SDHI_INFO2);
			if (stat & 0x100)
				break;
			if (stat & 0x7f)
				dev_err("status = %p\n", stat);
		} while (1);

		SDHI_WRITE16(host->base, SDHI_INFO2, 0xfeff);
		val = SDHI_READ16(host->base, SDHI_EXT_ACC);

		if (get_sdhi_port(host->base) == 0)
			SDHI_WRITE16(host->base,
				     SDHI_EXT_ACC, val | (1 << 0));
		else 
			SDHI_WRITE16(host->base,
				     SDHI_EXT_ACC, val & ~(1 << 0));

		while (blksz > 0) {
			*(buffer++) = SDHI_READ16(host->base, SDHI_BUF);
			blksz -= 2;
		}
	}

	return 0;
}

static int sdhc_do_cmd(struct mmc_host *host, struct mmc_command *cmd)
{
	u16 val;
	u32 mode;
	int i;
	u32 stat, stat2;
	static int acmd = 0;

	trace();

	if (cmd->opcode == MMC_GO_IDLE_STATE) {
		val = SDHI_READ16(host->base, SDHI_CLK_CTRL);
		val |= 0x0100;
		val &= ~0x0200;
		SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);
		udelay(10*1000);
		SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);
	}

	sdhc_clock_start(host);

	/* Waiting for SD Bus busy to be cleared */
	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));

	if (cmd->debug)
		dev_dbg("CMD%d\n", cmd->opcode);

	mode = cmd->opcode;
	if (acmd)
		mode |= (1 << 6);

	if (cmd->opcode == 55)
		acmd = 1;
	else
		acmd = 0;

	if (cmd->data) {
		mode |= (1 << 11);
		if (cmd->data->flags & MMC_DATA_READ)
			mode |= (1 << 12);

		if (cmd->data->blocks > 1) {
			mode |= (1 << 13);
			SDHI_WRITE16(host->base, SDHI_STOP, 0x100);

			SDHI_WRITE16(host->base, SDHI_SECCNT,
				     cmd->data->blocks);
			SDHI_WRITE16(host->base, SDHI_SIZE, 512);
		} else {
			SDHI_WRITE16(host->base, SDHI_SIZE, cmd->data->blksz);
		}
	}

	/* status clear */
	SDHI_WRITE16(host->base, SDHI_INFO1, 0x0000);
	SDHI_WRITE16(host->base, SDHI_INFO2, 0x0000);

	SDHI_WRITE16(host->base, SDHI_ARG0, cmd->arg & 0xffff);
	SDHI_WRITE16(host->base, SDHI_ARG1, (cmd->arg >> 16) & 0xffff);
	if (cmd->debug)
		dev_dbg("XFERTYP: %p\n", mode);

	SDHI_WRITE16(host->base, SDHI_CMD, mode);


	i=1000;
	if (cmd->opcode != 12) {
		for (i=1000000; i; i--) {
			stat = SDHI_READ16(host->base, SDHI_INFO1);
			if (stat & 0x1)
				break;
			udelay(1);
		}
	}
	SDHI_WRITE16(host->base, SDHI_INFO1, 0xfffe);

	stat2 = SDHI_READ16(host->base, SDHI_INFO2);
	if ((i <= 0) || (stat2 & 0x7f)) {
		if (cmd->opcode != SD_SEND_IF_COND)
			dev_err("CMD:%d timedout. (status = %p)\n",
				cmd->opcode, stat2, i);
		return -H_EIO;
	}

	if (cmd->flags & MMC_RSP_136) {
		u32 rsp[8];

		for (i=0; i<8; i++)
			rsp[i] = (u32)SDHI_READ16(host->base,
						  (SDHI_RSP0 + 2 * i));

		cmd->resp[3] = (((rsp[1] << 24) & 0xff000000) |
				((rsp[0] << 8) & 0xff00));
		cmd->resp[2] = (((rsp[3] << 24) & 0xff000000) |
				(rsp[2] << 8) | ((rsp[1] >> 8) & 0x00ff));
		cmd->resp[1] = (((rsp[5] << 24) & 0xff000000) |
				(rsp[4] << 8) | ((rsp[3] >> 8) & 0x00ff));
		cmd->resp[0] = (((rsp[7] << 24) & 0xff000000) |
				(rsp[6] << 8) | ((rsp[5] >> 8) & 0x00ff));
	} else {
		cmd->resp[0] = ((SDHI_READ16(host->base, SDHI_RSP1) << 16) |
				SDHI_READ16(host->base, SDHI_RSP0));
	}

	if (cmd->debug)
		dev_dbg("RESP: %p, %p, %p, %p\n",
			cmd->resp[0], cmd->resp[1],
			cmd->resp[2], cmd->resp[3]);

	if (cmd->data) {
		sdhc_pio_transfer_data(host, cmd);

		if (cmd->data->stop) {
			int ret;
			ret = sdhc_do_cmd(host, cmd->data->stop);
			if (ret)
				return ret;
		}
	}

	return 0;
}

static int sdhc_get_card_detect(struct mmc_host *host)
{
	u16 val;

	val = SDHI_READ16(host->base, SDHI_INFO1);
	if (val & 0x20)
		return 1; /* Card Detected */

	return 0;
}

static int sdhc_set_bus_width(struct mmc_host *host, u32 width)
{
	u32 val;

	trace();

	val = SDHI_READ16(host->base, SDHI_OPTION);

	switch (width) {
	case 4:
		val &= ~(1 << 15);
		break;
	default:
		val |= (1 << 15);
		break;
	}
	SDHI_WRITE16(host->base, SDHI_OPTION, val);

	return 0;
}

static void sdhc_reset(struct mmc_host *host)
{
	trace();

	udelay(1*1000);
}

static void sdhc_cpg_enable(int port)
{
	u32 sr, cr;
	u32 bit;

	if (port == 1)
		bit = CPG_MSTP3_SDHI1;
	else
		bit = CPG_MSTP3_SDHI0;

	sr = read32(CPG_MSTPSR3);
	if (sr & bit) {
		cr = read32(CPG_SMSTPCR3);
		write32(CPG_SMSTPCR3, cr & ~bit);
	}

	while (read32(CPG_MSTPSR3) & bit);
}

static void sdhi_set_cd_detect_time(struct mmc_host *host)
{
	u16 val;
	int port;

	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));

	val = SDHI_READ16(host->base, SDHI_OPTION);
	val &= ~0xf;

	/* about 1.6msec */
	port = get_sdhi_port(host->base);
	if (port == 0)
		val |= 0x4;
	else
		val |= 0x3;

	SDHI_WRITE16(host->base, SDHI_OPTION, val);
}

static void __maybe_unused sdhi_voltage_init(void)
{
	u8 buf;

	buf = 0;
	da9063_write(0x28, &buf, 1);
	da9063_write(0x29, &buf, 1);

	udelay(1 * 1000);

	gpio_direction_output(xSD0_PWR_EN, 0);
	gpio_direction_output(xSD1_PWR_EN, 0);

	udelay(100 * 1000);

	gpio_direction_output(xSD0_PWR_EN, 1);
	gpio_direction_output(xSD1_PWR_EN, 1);

	udelay(5 * 1000);

	buf = 0;
	da9063_write(0x32, &buf, 1);
	da9063_write(0x33, &buf, 1);

	buf = 1;
	da9063_write(0x28, &buf, 1);
	da9063_write(0x29, &buf, 1);

	udelay(300);
}

static int sdhc_init(struct mmc_host *host)
{
	int port;

#if defined(CONFIG_PLATFORM_ARMADILLOEVA1500)
	sdhi_voltage_init();
#endif

	trace();

	port = get_sdhi_port(host->base);

	rcm2_sdhc_request_port(port);

	sdhc_set_pwr(port, SDHC_POWER_OFF);

	sdhc_cpg_enable(port);

	sdhc_reset(host);

	udelay(200 * 1000);
	sdhc_set_pwr(port, SDHC_POWER_ON);
	udelay(10 * 1000);

	SDHI_WRITE16(host->base, SDHI_SOFT_RST, 0x1);

	sdhi_set_cd_detect_time(host);
	udelay(5 * 1000);

	sdhc_set_clock(host, 0);
	sdhc_set_bus_width(host, 1);

	return 0;
}

static void sdhc_exit(struct mmc_host *host)
{
	trace();
}

static struct mmc_host_ops rcm2_sdhc_ops = {
	.get_cd = sdhc_get_card_detect,
	.do_cmd = sdhc_do_cmd,
	.set_clock = sdhc_set_clock,
	.set_bus_width = sdhc_set_bus_width,
};

struct mmc_host rcm2_sdhc = {
	.init = sdhc_init,
	.exit = sdhc_exit,
	.ops = &rcm2_sdhc_ops,
};
